/*
 * Copyright (c) 2019 Chris Camacho (codifies -  http://bedroomcoders.co.uk/)
 * April '22 updated for latest release of Raylib
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stddef.h>

#include "raylib.h"
#include "raymath.h"

#define RLIGHTS_IMPLEMENTATION
#include "rlights.h"

#define screenWidth 1280
#define screenHeight 720

#include "map.h"

const int map[MAPWIDTH][MAPHEIGHT] = {
    {2,2,2,0,2,2,2,2,0,2,2,0,0,2,2,2},
    {2,2,0,0,0,2,2,0,0,0,2,0,0,0,0,2},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,3,3,3,3,3,0,0,0,0,0},
    {0,0,0,0,3,3,3,3,3,3,3,3,0,0,0,0},
    {0,0,0,3,3,3,3,3,3,3,3,3,3,0,0,0},
    {0,0,3,3,3,3,3,3,3,3,3,3,3,0,0,0},
    {3,4,3,3,3,3,3,3,3,3,3,3,3,0,0,0},
    {0,0,3,3,3,3,3,3,3,3,3,3,3,0,0,0},
    {0,0,0,3,3,3,3,3,3,3,3,3,3,0,0,0},
    {0,0,0,0,3,3,3,3,3,3,3,3,0,0,0,0},
    {0,0,0,0,0,0,3,3,3,3,3,0,0,0,0,0},
    {0,0,0,1,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {0,0,0,0,0,0,0,0,0,0,0,1,0,0,0,0},
    {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0},
    {2,0,0,0,0,2,2,0,0,0,0,0,0,0,0,2},
    {2,2,0,0,2,2,2,2,2,2,0,2,0,0,2,2}
};


const float hsize = 1.004;
const float vsize = 1.008;


Vector2 mouse2hex(Vector2 m) {
    float X = ((((1.73205080757)/3*m.x ) / hsize))+hsize/2;
    float Z = ((2./3*m.y) / vsize)+vsize/2;
    if ((int)Z & 1) X-=.5;
    return (Vector2){(int)X,(int)Z};
}

int main(void)
{
    // TODO "load" the mapNodes from map array but should be from a file...
    for (int y=0; y<MAPHEIGHT; y++) {
        for (int x=0; x<MAPWIDTH; x++) {
            mapNodes[x][y].type = map[x][y];
            mapNodes[x][y].coords.x=x;
            mapNodes[x][y].coords.y=y;
        }
    }

    // builds the "graph" using our terrain rules
    buildMapLinks();

    // Initialization
    //--------------------------------------------------------------------------------------
    InitWindow(screenWidth, screenHeight, "raylib - test");

    // Define the camera to look into our 3d world
    Camera camera = { 0 };
    camera.position = (Vector3){ 16.0f, 16.0f, -16.0f };
    camera.target = (Vector3){ 16.0f, 0.0f, 16.0f };
    camera.up = (Vector3){ 0.0f, 1.0f, 0.0f };
    camera.fovy = 45.0f;
    camera.projection = CAMERA_PERSPECTIVE;
    SetCameraMode(camera, CAMERA_FREE);
    SetCameraAltControl(KEY_SPACE);                       // Set camera alt key to combine with mouse movement (free camera)
    SetCameraPanControl(MOUSE_LEFT_BUTTON);

    // lighting shader
    Shader shader = LoadShader("data/simpleLight.vs", "data/simpleLight.fs");
    shader.locs[SHADER_LOC_MATRIX_MODEL] = GetShaderLocation(shader, "matModel");
    shader.locs[SHADER_LOC_VECTOR_VIEW] = GetShaderLocation(shader, "viewPos");

    // ambient light level
    int amb = GetShaderLocation(shader, "ambient");
    SetShaderValue(shader, amb, (float[4]){0.2,0.2,0.2,1.0}, SHADER_UNIFORM_VEC4);

    Texture boulderTx, bridgeTx, treeTx, grassTx, waterTx;
    Model boulder, hex, bridge, tree, pointer;
    boulderTx = LoadTexture("data/boulderHex.png");
    bridgeTx = LoadTexture("data/bridge.png");
    treeTx = LoadTexture("data/treeHex.png");
    grassTx = LoadTexture("data/grass.png");
    waterTx = LoadTexture("data/water.png");

    boulder = LoadModel("data/boulderhex.obj");
    boulder.materials[0].shader = shader;
    boulder.materials[0].maps[MATERIAL_MAP_DIFFUSE].texture = boulderTx;

    // the hex model is retextured and repurposed for those terrains
    // that are simply hex shaped grass, water and similar flat ones
    hex = LoadModel("data/hex.obj");
    hex.materials[0].shader = shader;

    bridge = LoadModel("data/bridgehex.obj");
    bridge.materials[0].shader = shader;
    bridge.materials[0].maps[MATERIAL_MAP_DIFFUSE].texture = bridgeTx;

    tree = LoadModel("data/treehex.obj");
    tree.materials[0].shader = shader;
    tree.materials[0].maps[MATERIAL_MAP_DIFFUSE].texture = treeTx;

    pointer = LoadModel("data/pointer.obj");


    // make a light (max 4 but we're only using 1)
    Light light = CreateLight(LIGHT_POINT, (Vector3){ 2,4,1 }, Vector3Zero(), WHITE, shader);

    // frame counter
    int frame = 0;

    SetTargetFPS(60);               // Set  to run at 60 frames-per-second
    HideCursor();

    mapNode_t* pathStart = NULL;
    mapNode_t* pathEnd = NULL;
    //--------------------------------------------------------------------------------------
    // Main game loop
    while (!WindowShouldClose())    // Detect window close button or ESC key
    {
        // Update
        //----------------------------------------------------------------------------------

        frame ++;

        UpdateCamera(&camera);

        // position the light slightly above the camera
        light.position = camera.position;
        light.position.y += 0.1f;

        // update the light shader with the camera view position
        SetShaderValue(shader, shader.locs[SHADER_LOC_VECTOR_VIEW], &camera.position.x, SHADER_UNIFORM_VEC3);
        UpdateLightValues(shader, light);

        // all the hexes are laying on a flat plane so we just project
        // a mouse ray onto this plane
        Ray r = GetMouseRay(GetMousePosition(), camera);
        RayCollision rhi = GetRayCollisionQuad(r, 
                                (const Vector3){-1000,0,-1000},
                                (const Vector3){ 1000,0,-1000},
                                (const Vector3){ 1000,0, 1000},
                                (const Vector3){-1000,0, 1000});

        // turn the plane coordinates into hex map coordinates
        Vector2 curs = mouse2hex((Vector2){rhi.point.x,rhi.point.z});

        if (inMap(curs)){
            if (IsKeyPressed(KEY_S)) {
                pathStart = &mapNodes[(int)curs.x][(int)curs.y];
                if (pathEnd) findPath(pathStart, pathEnd);
            }
            if (IsKeyPressed(KEY_E)) {
                pathEnd = &mapNodes[(int)curs.x][(int)curs.y];
                if (pathStart) findPath(pathStart, pathEnd);
            }
            if (IsKeyPressed(KEY_T)) {
                mapNode_t* m = &mapNodes[(int)curs.x][(int)curs.y];
                m->type++;
                if (m->type > HxBridge) {
                    // above highest terrain type so wrap around
                    m->type = 0;
                }
                // we need only rebuild map links for the hex changed and
                // its neighbours
                buildMapLinks();
                if (pathStart && pathEnd) {
                    findPath(pathStart, pathEnd);
                }
            }
        }



        // Draw
        //----------------------------------------------------------------------------------
        BeginDrawing();

            ClearBackground(BLACK);

            BeginMode3D(camera);

                for (int y=0;y<MAPHEIGHT;y++) {
                    for (int x=0;x<MAPWIDTH;x++) {
                        float offset = 0;
                        // odd rows are offset...
                        if (y/2 != (float)y/2.0) offset = OFFSET;
                        Model* m = NULL;
                        int tp = mapNodes[x][y].type;
                        // select the right model for the terrain
                        if (tp==HxGrass) {
                            m = &hex;
                            m->materials[0].maps[MATERIAL_MAP_DIFFUSE].texture = grassTx;
                        }
                        if (tp==HxBoulder) m = &boulder;
                        if (tp==HxTree) m = &tree;
                        if (tp==HxWater) {
                            m = &hex;
                            m->materials[0].maps[MATERIAL_MAP_DIFFUSE].texture = waterTx;
                        }
                        if (tp==HxBridge) m = &bridge;
                        if (m) {
                            // change the tint depending on path and highlighted
                            // hexes
                            Color tint = WHITE;
                            if (x==(int)curs.x && y==(int)curs.y) tint = RED;
                            if (mapNodes[x][y].onPath) tint = ORANGE;
                            if (pathStart == &mapNodes[x][y]) tint = MAGENTA;
                            if (pathEnd == &mapNodes[x][y]) tint = GREEN;
                            DrawModel(*m,(Vector3){x*HWIDTH+offset,0,y*HHEIGHT},1,tint);
                        }
                    }
                }

/* testing projected mouse to hex
                for (float y=-1;y<MAPHEIGHT*HHEIGHT;y+=0.2) {
                    for (float x=-1;x<MAPWIDTH*HWIDTH;x+=0.2) {

                        Vector2 p = mouse2hex((Vector2){x,y});
                        Color tint = WHITE;
                        if ( (int)p.x & 1 ) tint.r = 0;
                        if ( (int)p.y & 1 ) tint.g = 0;
                        DrawCube((Vector3){x,0.2,y},0.1,0.1,0.1,tint);
                    }
                }
*/

                DrawModel(pointer, rhi.point, 1, WHITE);

            EndMode3D();

            DrawFPS(10, 10);

            int l = MeasureText(TextFormat("Frame %i", frame), 20);
            DrawRectangle(16, 698, l+8, 42, BLUE);
            DrawText(TextFormat("Frame %i", frame), 20, 700, 20, WHITE);
            DrawText("S, E to place path start and end points", 20, 40, 20, GREEN);
            DrawText("T to change the tile type under the cursor", 20, 60, 20, GREEN);
            DrawText("Drag mouse to move camera", 20, 80, 20, GREEN);
            DrawText("Hold space and drag the mouse to rotate the camera", 20, 100, 20, GREEN);
            DrawText("Mouse wheel to zoom", 20, 120, 20, GREEN);

        EndDrawing();
        //----------------------------------------------------------------------------------
    }

    // De-Initialization
    //--------------------------------------------------------------------------------------

    UnloadModel(boulder);
    UnloadModel(hex);
    UnloadModel(bridge);
    UnloadModel(tree);
    UnloadModel(pointer);
    UnloadTexture(boulderTx);
    UnloadTexture(bridgeTx);
    UnloadTexture(treeTx);
    UnloadTexture(grassTx);
    UnloadTexture(waterTx);

    UnloadShader(shader);

    CloseWindow();        // Close window and OpenGL context
    //--------------------------------------------------------------------------------------

    return 0;
}



