/*
 * Copyright (c) 2019 Chris Camacho (codifies -  http://bedroomcoders.co.uk/)
 * April '22 updated for latest release of Raylib
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdlib.h> // abs / rand
#include "raylib.h"
#include "map.h"
#include "clist.h"

point_t evenDir[6] = { {1,0},{1,1},{0,1},{-1,0},{0,-1},{1,-1} };
point_t oddDir[6] = { {1,0},{0,1},{-1,1},{-1,0},{-1,-1},{0,-1} };

mapNode_t mapNodes[MAPWIDTH][MAPHEIGHT];


bool inMap(Vector2 c) {
    if (c.x < 0 || c.y < 0) return false;
    if (c.x > MAPWIDTH-1 || c.y > MAPHEIGHT-1) return false;
    return true;
}

void buildMapLinks()
{
    for (int y=0; y<MAPHEIGHT; y++)
    {
        for (int x=0; x<MAPWIDTH; x++) {
            mapNodes[x][y].onPath = 0;
            mapNodes[x][y].H = 999;
            mapNodes[x][y].G = 999;
            mapNodes[x][y].passable = 0;
            int m = mapNodes[x][y].type;
            if (m==HxGrass || m==HxBridge) {
                mapNodes[x][y].passable = 1;

                for (int n=0; n<6; n++) {
                    mapNodes[x][y].exits[n]=0;
                    int mm,xx,yy;
                    if ((int)y % 2 != 0) {
                        yy=y+evenDir[n].y;
                        xx=x+evenDir[n].x;
                    } else {
                        yy=y+oddDir[n].y;
                        xx=x+oddDir[n].x;
                    }
                    if (inMap((Vector2) {xx,yy})) {
                        mm = mapNodes[xx][yy].type;
                        if (mm==HxGrass || mm==HxBridge) {
                            mapNodes[x][y].exits[n] = &mapNodes[xx][yy];
                        }
                        if (m==HxBridge || mm==HxBridge) {
                            if (n!=0 && n!=3) {
                                mapNodes[x][y].exits[n] = 0;
                            }
                        }
                    }
                }
            } else {
                mapNodes[x][y].passable = 0;
            }
        }
    }
}

float distance(int ax, int ay, int bx, int by)
{
    int cax = ay;
    int caz = ax - (ay + (ay&1)) / 2.f;
    int cay = -cax - caz;
    int cbx = by;
    int cbz = bx - (by + (by&1)) / 2.f;
    int cby = -cbx - cbz;
    return (abs(cax - cbx) + abs(cay - cby) + abs(caz - cbz)) / 2.f;
}

// compare function for sorting the open list
int cmpNodes(cnode_t* n1, cnode_t* n2)
{
    mapNode_t* rn1 = (mapNode_t*)(n1->data);
    mapNode_t* rn2 = (mapNode_t*)(n2->data);
    if (rn1->G + rn1->H > rn2->G + rn2->H) return 1;
    return 0;
}

// reset the map ready for new path discovery
void resetPath()
{
    for (int y=0; y<MAPHEIGHT; y++)
    {
        for (int x=0; x<MAPWIDTH; x++)
        {
            mapNodes[x][y].onPath = 0;
            mapNodes[x][y].H = 999;
            mapNodes[x][y].G = 999;
            mapNodes[x][y].pathParent = NULL;
        }
    }
}

void findPath(mapNode_t* from, mapNode_t* to)
{
    resetPath();
    clist_t *openList = clistCreateList();
    clist_t *closedList = clistCreateList();
    clistAddNode(openList, from);
    from->G=2;
    from->H=distance(from->coords.x,from->coords.y,to->coords.x,to->coords.y);
    while(!clistIsEmpty(openList)) {
        clistSort(openList, cmpNodes);
        // randomise top ties - equally short alterative paths chosen at random
        mapNode_t* current = openList->head->data;
        if (openList->head->next!=NULL) {
            mapNode_t* nn = openList->head->next->data;
            if (current->G+current->H == nn->G+nn->H ) {
                if (rand()/(float)RAND_MAX > 0.5) {
                    current = nn;
                }
            }
        }
        if ( current == to) break;

        // check all six possible exits
        for (int d=0; d<6; d++) {
            if (!current->exits[d]) continue;
            mapNode_t* successor = current->exits[d];
            if (!clistFindNode(closedList,successor) && successor->G > current->G+2) {
                if (clistFindNode(closedList,successor)!=NULL) {
                    clistDeleteNodeFromData(closedList,successor);
                }
                if (clistFindNode(openList,successor)!=NULL) {
                    clistDeleteNodeFromData(openList,successor);
                }
                if (successor->passable) {
                    if (successor->G > current->G+2) {
                        successor->G = current->G+2;
                        successor->pathParent = current;
                        successor->H = distance(current->coords.x,current->coords.y,to->coords.x,to->coords.y);
                        clistAddNode(openList, successor);
                    } else {
                        clistAddNode(closedList,successor);
                    }
                }
            }

        }

        clistAddNode(closedList,current);
        clistDeleteNodeFromData(openList,current);
    }

    // reverse through parent pointers marking path
    // only works if there is a complete path
    mapNode_t* n = to;
    while(n->pathParent) {
        n->onPath=1;
        n=n->pathParent;
    }
    from->onPath=1;

    // where a complete path can't be made, find the nearest
    // point to the end we could access
    // work the path back from that point
    if (to->pathParent == NULL) {
        int bestv = 999;
        mapNode_t* best = NULL;
        for (int y=0; y<MAPHEIGHT; y++) {
            for (int x=0; x<MAPWIDTH; x++) {
                mapNode_t* m = &mapNodes[x][y];
                if (m->H < bestv) {
                    bestv = m->H;
                    best = m;
                }
            }
        }
        if (best) {
            mapNode_t* n = best;
            while(n->pathParent) {
                n->onPath=1;
                n=n->pathParent;
            }
        }
    }

    clistFreeList(&openList);
    clistFreeList(&closedList);
    // TODO it might be nice to return a list of the path
    // rather than marking the path on the actual map
}
